#ifndef EL__PROTOCOL_FILE_MAILCAP_H
#define EL__PROTOCOL_FILE_MAILCAP_H

#include "main/module.h"
#include "protocol/protocol.h"

extern struct module mailcap_protocol_module;
extern protocol_handler_T mailcap_protocol_handler;

#endif
